
from langchain_core.prompts import ChatPromptTemplate


def prepare_prompt():
    # format prompt
    prompt = ChatPromptTemplate.from_template(
        """Answer the question based only on the following context: {context} Question: {question} """
    )
    return prompt


def get_perspectives_prompt():
    prompt = ChatPromptTemplate.from_template("""You are an AI language 
    model assistant. Your task is to generate five different versions of the 
    given user question to retrieve relevant documents from a vector database. 
    By generating multiple perspectives on the user question, your goal is to 
    help the user overcome some of the limitations of the distance-based 
    similarity search. Provide these alternative questions separated by 
    newlines. Original question: {question}""")

    return prompt


def get_fusion_rag_prompt():

    prompt = ChatPromptTemplate.from_template(
        """You are a helpful assistant that generates multiple search queries 
        based on a single input query. \n Generate multiple search queries 
        related to: {question} \n Output (4 queries):""")
    return prompt


def get_hyde_prompt():
    prompt = ChatPromptTemplate.from_template(
        """Please write a passage to answer the question.\n Question: {question} \n Passage:""")
    return prompt



from langchain_community.document_loaders import TextLoader
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain_postgres.vectorstores import PGVector
from core.utils import get_connection_string, get_embeddings_model


def read_document(file_path):
    loader = TextLoader(file_path)
    return loader.load()


def process_chunks(file_path):
    raw_documents = read_document(file_path)
    text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=1000, chunk_overlap=200)
    return text_splitter.split_documents(raw_documents)


def create_embeddings_and_upsert(file_path, collection_name="raptai"):
    documents = process_chunks(file_path)
    connection = get_connection_string()
    embeddings_model = get_embeddings_model()
    return PGVector.from_documents(
        documents, embeddings_model, connection=connection, collection_name=collection_name)





from langchain_postgres import PGVector
from core.utils import get_llm
from services.prompt_service import get_hyde_prompt, get_perspectives_prompt, get_fusion_rag_prompt
from services.retrieval_service import parse_queries_output
from langchain_core.output_parsers import StrOutputParser


def generate_query():
    prompt = get_perspectives_prompt()
    llm = get_llm()

    query_gen = prompt | llm | parse_queries_output
    return query_gen


def generate_fusion_query():
    prompt = get_fusion_rag_prompt()
    llm = get_llm()

    query_gen = prompt | llm | parse_queries_output
    return query_gen


def generate_hyde_query():
    prompt = get_hyde_prompt()
    llm = get_llm()

    query_gen = prompt | llm | StrOutputParser()
    return query_gen




from langchain_postgres.vectorstores import PGVector
from core.document_utils import get_unique_union, reciprocal_rank_fusion
from core.utils import get_connection_string, get_embeddings_model


def get_vectorstore(collection_name):
    vectorstore = PGVector.from_existing_index(
        collection_name=collection_name,
        embedding=get_embeddings_model(),
        connection=get_connection_string()
    )
    return vectorstore


def create_retriever(collection_name):
    db = get_vectorstore(collection_name)
    # create retriever to retrieve 2 relevant document
    return db.as_retriever(search_kwargs={"k": 2})


def parse_queries_output(message):
    return message.content.split('\n')


def retrieve_relevant_documents(collection_name, query: str):
    retriever = create_retriever(collection_name)

    # fetch relevant documents
    docs = retriever.invoke(query)
    return docs


def get_retriever_chain(collection_name, query_gen):
    retriever = create_retriever(collection_name)

    retrieval_chain = query_gen | retriever.batch | get_unique_union
    return retrieval_chain


def get_fusion_retriever_chain(collection_name, query_gen):
    retriever = create_retriever(collection_name)

    retrieval_chain = query_gen | retriever.batch | reciprocal_rank_fusion
    return retrieval_chain


def get_hyde_retriever_chain(collection_name, query_gen):
    retriever = create_retriever(collection_name)

    retrieval_chain = query_gen | retriever
    return retrieval_chain




from langchain_openai import OpenAIEmbeddings
from langchain_openai import ChatOpenAI
import os
from dotenv import load_dotenv

load_dotenv()


def get_connection_string():
    return os.environ.get("CONNECTION_STRING")


def get_embeddings_model():
    return OpenAIEmbeddings(api_key=os.environ.get("OPENAI_API_KEY"))


def get_llm():
    return ChatOpenAI(model_name="gpt-3.5-turbo", temperature=0, api_key=os.environ.get("OPENAI_API_KEY"))



def get_unique_union(document_lists):
    # Flatten list of lists, and dedupe them
    deduped_docs = {
        doc.page_content: doc for sublist in document_lists for doc in sublist}
    # return a flat list of unique docs
    return list(deduped_docs.values())


def reciprocal_rank_fusion(results: list[list], k=60):
    """reciprocal rank fusion on multiple lists of ranked documents and an optional parameter k used in the RRF formula"""
    # Initialize a dictionary to hold fused scores for each document
    # Documents will be keyed by their contents to ensure uniqueness
    fused_scores = {}
    documents = {}
    for docs in results:
        # Iterate through each document in the list, with its rank (position in the list)
        for rank, doc in enumerate(docs):
            doc_str = doc.page_content
            if doc_str not in fused_scores:
                fused_scores[doc_str] = 0
                documents[doc_str] = doc
            fused_scores[doc_str] += 1 / (rank + k)
    # sort the documents based on their fused scores in descending order to get the final reranked results
    reranked_doc_strs = sorted(
        fused_scores, key=lambda d: fused_scores[d], reverse=True)
    return [documents[doc_str] for doc_str in reranked_doc_strs]



from langchain_core.runnables import chain
from core.utils import get_llm
from services.generation_service import generate_fusion_query
from services.prompt_service import prepare_prompt
from services.retrieval_service import get_fusion_retriever_chain


def user_query():
    return 'Difference between quantum computing and classical computing?'


collection_name = 'quantum_computing'


@chain
def qa(input):
    query_gen = generate_fusion_query()
    retrieval_chain = get_fusion_retriever_chain(collection_name, query_gen)

    docs = retrieval_chain.invoke(input)

    # prepare prompt
    prompt = prepare_prompt()
    # format prompt
    formatted = prompt.invoke({"context": docs, "question": input})

    llm = get_llm()
    # generate answer
    answer = llm.invoke(formatted)
    return answer


# run it
result = qa.invoke(user_query())
print(result.content)
